/*	Pace Clock on ATmega644
 *	ECE 4760 Spring 2011
 *	Final Project
 *
 * 	By Paul Swirhun and Andy Liang
 */

//TODO: a lot of optimization (remove a lot of _delay_us and multiple GSCLK pulses)

/*	Used to interface AVR microcontroller with
 *	TLC5940 LED driver chip from Texas Instruments
 */

#include "LEDinterface.h"
#include <avr/io.h>
#include <util/delay.h>

#define begin {
#define end }

//Gives 7-seg outputs indexed with digit 0..9,ALL_OFF,"R","C","P"
//However, bit 7 is left as 1 to enable colon/hyphen decoding
unsigned char seg7tbl[24] = {0b11111110,	//0
							 0b10110000,	//1
							 0b11101101,	//2
							 0b11111001,	//3
							 0b10110011,	//4
							 0b11011011,	//5
							 0b11011111,	//6
							 0b11110000,	//7
							 0b11111111,	//8
							 0b11110011,	//9
							 0b10000000,	//*OFF*
							 0b11000110,	//R
							 0b11001110,	//C
							 0b11100111,	//P
							 0b10001110,	//L
							 0b10110111,	//H
							 0b11111011,	//G
							 0b10011101,	//o
							 0b10111110,	//U
							 0b11000111,	//F
							 0b11110111,	//A
							 0b10010101,	//n
							 0b10111101,	//d
							 0b11001111		//E
							 };	

//Variables used to interface with the LED display
volatile unsigned char LED_buffer_long[96]; //used to send data to the four LED controllers
unsigned char LED_buffer_dot_correction_long[48]; //dot correction data
unsigned int brightness; //used to store brightness level

//**********************************************************
/*	Send k bytes to LED controllers, starting with index (k-1) and going to (0)
	Sends MSB first (ie, from index 95 to 0). k is the number of bytes to send,
	and must be less than 127 in this implementation. Note that each TLC5940
	in a chain of them uses 24 bytes, so we will typically have k==96.
	The function latches the data in, pulses BLANK to reset counter, and 
	pulses GSCLK to give DC illumination.
 */
void send_TLC5940(unsigned char vals[], unsigned char k)
begin
	signed char i = k;
	//Simply write to SPDR to initiate data transmission
	for (i -= 1; i >= 0; i--) {
		SPDR = vals[i]; //send value over serial MOSI

		//SPIF bit in SPSR is set when serial transfer is complete
		while (!(SPSR & (1<<SPIF))); //wait for completion
	}
	SPSR = SPSR & (~(1<<SPIF)); //turn off the SPIF flag
	//Latch the data with a pulse on XLAT
	LED_PORT |= (1 << XLAT); //turn on XLAT
	LED_PORT &= (~(1 << XLAT)); //turn off XLAT
	//Pulse BLANK to make sure the TLC5940 counter is reset
	LED_PORT |= (1 << BLANK);
	LED_PORT &= (~(1 << BLANK));
	//Give a rising edge on GSCLK to increment counter once
	LED_PORT |= (1 << GSCLK);
	LED_PORT &= (~(1 << GSCLK));
end

//**********************************************************
//Sets up the SPI interface for the LED controllers
void LED_init()
begin
	LED_PORT = (1<<MISO); //enable pullup on MISO (unused port).
	//Set the following communication pins to outputs
	LED_DDR = ((1<<SCK) | (1<<MOSI) | (1<<BLANK) | (1<<VPRG) | (1<<GSCLK) | (1<<XLAT));


	//Enable SPI, MSB first, Master, //low clock when idle,
	//sample on leading edge, use fosc/2 transmission rate (fastest)
	SPCR = ((1<<SPE) | (1<<MSTR)); //SPR1,SPR0 = 0 for clk/4 or clk/2 operation
	SPSR = (1<<SPI2X); //turn on SPI 2x speed enhancement bit to use clk/2
	

	//Set up the brightness, dot correction
	brightness = 0x0FFF; //set brightness to max (4095)
	setup_LED_buffer_dot_correction(); //initializes dot correction vector
	
	//Initialize the dot-correction on the LED drivers
	LED_PORT |= (1<<VPRG); //turn on VPRG for dot-correction (sent once);
	send_TLC5940(LED_buffer_dot_correction_long,48); //send dot correction data
	//Switch to the grayscale mode (sent every time to update the LED display)
	LED_PORT &= (~(1<<VPRG));

end


//**********************************************************
/*	Overwrites the LED_buffer_long, a 96 byte dataset to send to LED controllers.
	This function calls setup_LED_buffer

	digits[4] holds the digits in a 1-hot ABCDEFG code.
	note that digits[0] is the rightmost display digit.
	to display a time of 12:30, digits[] = {0,3,2,1}

	This function also applies the colon, hyphen, and period settings.
	
	If null_high ==1, then the 2 most significant panels are turned off
	if it won't change the time displayed on screen. This means digits[3] is
	turned off if it is zero, and digits[2] is turned off if digits[3] and
	digits[2] are both zero.
	
	If null_high ==2, then the 1 most siginficant panel is turned off
	if it won't change the time displayed on screen. This means digits[3] is
	turned off if it is zero.
	
	If null_high ==0, then no panels are turned off and they display the inputs.
	
	This is done by copying out to a local variable,
	so the input data array will not be modified. */
void write_LED_buffer(unsigned char digits[4], unsigned char colon, unsigned char null_high)
begin
	unsigned char digits_copy[4]; //stores copy with higher panel selectively nulled
	digits_copy[3] = digits[3]; //may be modified
	digits_copy[2] = digits[2];	//may be modified
	digits_copy[1] = digits[1]; //least significant 2 panels are not modified
	digits_copy[0] = digits[0];

	if(null_high == 1) { //then calling function requested the upper 2 panels be selectively turned off
		if(digits_copy[3] == 0) {
			digits_copy[3] = 10; //turn most significant panel off
			if(digits_copy[2] == 0)
				digits_copy[2] = 10; //turn second most significant panel off
		}
	} else if((null_high == 2) && (digits_copy[3] == 0))
		digits_copy[3] = 10; //turn most significant panel off

	setup_LED_buffer(); //set up with appropriate brightness
	
	unsigned char n,d,k,b,d7seg;
	d = 0; //index into digits_copy[]

	for(n=0; n < 96; n = n+24) { //loop through the four digits
		
		d7seg = seg7tbl[digits_copy[d]]; //decode digit into 7seg format
		d++; //increment to next digit to display		

		b = 0; //bit index to set each pair of octets

		for(k=n; k < (n+24); k=k+3) {
			if(!((1<<b) & d7seg)) { //check if bit b of this digit should be off
				LED_buffer_long[k] = 0;
				LED_buffer_long[k+1] = 0;
				LED_buffer_long[k+2] = 0;
			}
			b++; //increment to the next bit of this digit
		}
	}

	//Selectively turn off colon/hyphen/period outputs here
	//lookup table should leave LED outputs 14 and 15 on
	b = 0;
	for(n=21;n<=93;n=n+24) {
		if (!((1<<b) & colon)) { //check if bit b of colon is off
			//Then zero out byte n and the lower half of byte n+1
			LED_buffer_long[n] = 0;
			LED_buffer_long[n+1] = (LED_buffer_long[n+1] & 0xF0);
		}

		b++; //go to the next bit in colon

		if (!((1<<b) & colon)) { //check if bit b of colon is off
			//Then zero out byte n+2 and the upper half of byte n+1
			LED_buffer_long[n+2] = 0;
			LED_buffer_long[n+1] = (LED_buffer_long[n+1] & 0x0F);
		}

		b++; //go to the next bit in colon
	}
end

//**********************************************************
/*	Overwrites the LED_buffer_long with 96 bytes of octet (12-bit) brightness values */
void setup_LED_buffer(void)
begin
	unsigned char top8 = (char)(brightness>>4); //top eight bits of brightness
	//top nibble is bottom nibble of brightness, bottom nibble is top nibble of brigtness
	unsigned char mid8 = (((char)(brightness<<4)) | ((char)(brightness>>8))); 
	unsigned char bot8 = (char)(brightness & 0x00FF); //bottom eight bits of brightness

	unsigned char k;
	for(k=0;k<96;k=k+3) {
		LED_buffer_long[k] 		= bot8;
		LED_buffer_long[k+1] 	= mid8;
		LED_buffer_long[k+2] 	= top8;
	}
end

//**********************************************************
/* Overwrites the LED_buffer_dot_correction vector (all equalized, at this time)*/
void setup_LED_buffer_dot_correction(void)
begin
	unsigned char k;
	for(k=0;k<48;k++)
		LED_buffer_dot_correction_long[k] = 0xFF;
end

