/*	Pace Clock on ATmega32A
 *	Cornell University ECE 4760 Spring 2011
 *	Final Project
 *	For Atmel ATMega32A at 11.0592MHz
 * 	By Paul Swirhun and Andy Liang
 */

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <util/delay.h>
#include <avr/eeprom.h>
#include <avr/sleep.h>
#include "uart_mega32.h"
#include "i2cmaster.h"
#include "clockInterface.h"
#include "LEDinterface.h"
#include "eeprom_data.h"

#define begin {
#define end }

//For intuitive indexing into seg7tbl lookup table:
#define CONST_OFF	10
#define CONST_R		11
#define CONST_C		12
#define CONST_P		13
#define CONST_L		14
#define CONST_H		15
#define CONST_G		16
#define CONST_O		17
#define CONST_U		18
#define CONST_F		19
#define CONST_A		20
#define CONST_N		21
#define CONST_D		22
#define CONST_E		23

//For different amounts of EEPROM usage, adjust array size here:
#define UPLOAD_SIZE 1024 //number of bytes in a page-file EEPROM upload
#define UPLOAD_TIMER 479 //number of 32Hz ticks before upload times out (479 = 15seconds)

//available special modes encountered in the instruction set
enum MODES {SCHEDULED_START_MODE, MILLISECOND_TIMER_MODE, PROGRAM_PAUSE, 
USER_PAUSE, SET_CLOCK_HOURS, SET_CLOCK_MINUTES, ROUTINE_PROMPT, UPLOAD, 
UPLOAD_FAILED, UPLOAD_SUCCESS, SLEEP_MODE, NORMAL}; 

FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);

//Global Variables
extern volatile unsigned char LED_buffer_long[96]; //the raw PWM data sent to 4 TLC5940 controllers
unsigned volatile char upload_buffer[UPLOAD_SIZE]; //buffers inputs from burst-upload bluetooth transmission
unsigned  char aTime[4]; //stores the time to display left..right = index 3..0)
unsigned  char zTime[4]; //stores the target time (start point if counting down, end point if counting up)
unsigned volatile int aTime_integer, zTime_integer; //stores aTime,zTime but as an integer # of seconds
unsigned char hours_starting_point, minutes_starting_point; //for digit_selection in clock/schedule-start settings
unsigned char colon; //stores the decimal-place, hyphen, and colon display settings (1=hot encoding)
volatile char nextTimeReady; //a flag to indicate (if nonzero) that the next timing/instruction is ready
volatile char nextScreenReady; //a flag to indicate (if nonzero) that the next screen is ready
volatile char byteReceived; //the most recent byte received over bluetooth
volatile char user_paused; //!=0 when user has pushed pause (play/pause button), reset to 0 when user resumes
unsigned int data_index; //the index into the EEPROM data memory for the next thing to parse
unsigned int upload_data_index; //used for indexing into the upload_buffer[] and saving uploaded data
volatile unsigned int upload_timeout_timer; //counts up for upload timeout detection at 32ticks/sec
unsigned char time_data_raw[2]; //used to store minutes and seconds data from PCF8563
unsigned int routine_addresses[32]; //stores parsed addresses of routines for fast jumping
unsigned int inner_loop_header, outer_loop_header; //store location to jump back to for inner/outer loops
unsigned int inner_loop_footer, outer_loop_footer; //store location to jump back to for parsing backwards
char inner_loop_register, outer_loop_register; //store # of cycles remaining for inner/outer loops
char inner_loop_active, outer_loop_active; //!=0 if the inner/outer loops are active (allows nesting)
char count_up; //!=0 if counting up; =0 if counting down the current timing process
unsigned char current_routine; //stores routine being executed (0..31) for routine selection starting point
volatile char do_not_update; //!=0 to flag that the ISR should not send to the LED drivers
volatile char digit_selection; //used for user selecting hours, minutes, and routine numbers on-screen
volatile char ticks; //counts down from 31 to turn 32Hz external hardware clock into 1Hz
volatile char blinks_done; //counter used for blinking and timing status displays (usually 1-3 seconds)
volatile char ticks_since_parse_backwards; //counts up to 255; reset at 0 when user pushes "prev"
volatile unsigned int millisecond_ticks; //counter used for counting to 10-millisecond blocks
volatile enum MODES mode; //the current mode of operation

//Function Declarations (for forward references). See implementations below for descriptions.
void init();
void set_aTime_Hours_and_Minutes(void);
void set_aTime_Minutes_and_Seconds(void);
void set_azTime_Values(char val1, char val0, unsigned char v[]);
void parse_routines(void);
char Times_Equal(void);
void Pressed_Prev_Button(char);
void Pressed_Next_Button(void);
void Pressed_Enter_Button(void);
void Pressed_Play_Pause_Button(void);					
void Pressed_Upload_Button(void);
void Pressed_Power_Button(void);
void parse_backwards(char);
static inline void write_and_send(unsigned char null_high);
static inline void spin(char ticks_to_spin);
static inline void azset(char set_val3, char set_val2, char set_val1, char set_val0, unsigned char v[]);
void prompt_for_hours(unsigned char starting_point_input);
void prompt_for_minutes(unsigned char starting_point_input);
void reset_variables(void);
void handle_user_pause(void);
void handle_user_sleep(void);
void handle_user_upload(void);

//**********************************************************
/*	External Interrupt from PCF8563 32Hz clock signal
    On the ATmega32A, PinD.2 is the INT0 interrupt input*/
ISR (INT0_vect)
begin
	blinks_done--; //decrement at every 32Hz interrupt (used for blinking at rate > 1Hz)
	upload_timeout_timer++; //increment at every 32Hz interrupt to timeout a stalled upload
	if(mode == UPLOAD) //don't need screen update or "prev" functionality in upload mode
		return;

	//Inrement ticks_since_parse_backwards up to 255
	if(ticks_since_parse_backwards != 255)
		ticks_since_parse_backwards++;

	if(ticks == 0) {//then we are at 1Hz cycle: update the screen
		ticks = 31; //reset software counter
		
		if(do_not_update == 0) { //conditionally update based on this flag
			send_TLC5940(LED_buffer_long,96); //send to the LED drivers
			nextScreenReady = 0; //signal that system must prepare the next screen
		}
			
		if(Times_Equal()) {
			nextTimeReady = 0; //signal that the system must parse next timing/instruction		
			
		}
	} else {
		ticks--; //decrement tick counter
	}
}

//**********************************************************
/*	Timer/Counter0 Compare-Match Interrupt. Used to count 10 milliseconds
	for the millisecond-timer mode. init() assumes 11.0592MHz crystal.
	This should get executed once every 0.01 seconds when enabled, to
	accurately update the 10-millisecond timer functionality.*/
ISR (TIMER0_COMP_vect)
begin
	send_TLC5940(LED_buffer_long,96);
	nextScreenReady = 0;
	aTime_integer++; //increment the time in 10-ms steps
end

//**********************************************************
/*  Called when the USART Receives a Byte from Bluetooth.
    Buffers the value and sets a flag that a value was received,
	for interpretting that value in normal program execution */
ISR(USART_RXC_vect) 
begin
	//Must read UCSRA before reading UDR
	if(!(UCSRA & (1<<RXC)))
		return; //cannot afford a spin loop in an ISR (could take too long)

	byteReceived = UDR; //read the received byte

	if(mode != UPLOAD) { //parse the command if not in upload mode
		switch(byteReceived) { //The 5 buttons the user may press
			case 0x70:	//"Prev" button p
				if(ticks_since_parse_backwards > 47) //haven't pushed "prev" in 1.5 seconds
					Pressed_Prev_Button(1); //go back two elements
				else
					Pressed_Prev_Button(2); //go back two elements
				ticks_since_parse_backwards = 0; //reset when user pushes prev
				break;
			case 0x6E:	//"Next" button n
				Pressed_Next_Button();
				break;
			case 0x65:	//"Enter" buttton e
				Pressed_Enter_Button();
				break;
			case 0x61:	//"Play/Pause" button a
				Pressed_Play_Pause_Button();					
				break;
			case 0x75:	//"Upload" button u
				//Initiate handshake to start upload, then upload
				Pressed_Upload_Button();
				break;
			case 0x71: //"Power" button q (sleeps the CPU and turns display off)
				Pressed_Power_Button();
				break;
			default:
				break;
		}
	} else { //in upload mode. write to the upload buffer.
		if(upload_timeout_timer > UPLOAD_TIMER)//upload took too long
			mode = UPLOAD_FAILED;
		//fprintf(stdout,"Ind[%d] = %d\n\r",upload_data_index,byteReceived);
		upload_buffer[upload_data_index++] = byteReceived; //store uploaded data
		if(upload_data_index == UPLOAD_SIZE)
			mode = UPLOAD_SUCCESS; //done uploading. lets upload-loop continue
	}
end

//**********************************************************
/*	Main program entry point.
	Handles scheduling of tasks, parsing and interpretation of the instruction set,
	updating the screen, interacting with the user, and communicating with the 
	other integrated circuits (PCF8563, four TLC5940, RN-42).*/
int main()
begin
	unsigned char byte0, byte1; //used to temporarily store EEPROM data
	
	init(); //initialize everything

	while(1) {
		//Handle the events generated by user button-pushing (not stored instructions)
		handle_user_pause(); //Check for user-pause condition and block until escape
		handle_user_sleep(); //Check for user-sleep condition and sleep until escape		
		handle_user_upload(); //Check for user-upload condition, receive/process uploaded data

		if(nextScreenReady == 0) { //some updating of aTime must be done
			if(nextTimeReady == 0) { //must parse new time/instruction from EEPROM
				byte0 = get_data(data_index++); //read a byte and increment data_index			
				if(data_index == UPLOAD_SIZE) //out of bounds...wrap around to 0
					data_index = 0;

				if(0x80 & byte0) { //top bit is 1 -> instruction format data
					do_not_update = 1; //don't update the screen at 1Hz rate for instructions

					char IF10 = (0x03 & (byte0 >> 5)); //extract the IF1,IF0 bits
					unsigned char M40 = (0x1F & byte0);//extract the M4:M0 bits

					if(IF10) { //IF1,IF0 are nonzero
						switch(IF10) { //this instruction is a jump or loop type instruction
							case 0x03: //jump instruction
								data_index = routine_addresses[M40];
								current_routine = M40+1; //M40 encodes 
								break;
							case 0x02: //outer loop header
								if(outer_loop_active == 0) { //active the outer loop
									//Display "LooP" on the screen for 0.5 seconds
									spin(7); //wait 0.25 seconds 
									colon = 0; //colon/hyphen off
									azset(CONST_L,CONST_O,CONST_O,CONST_P,aTime);
									write_and_send(1);
									spin(15); //wait half a second

									outer_loop_register = M40+1; //number of iterations
									outer_loop_header = data_index; //record header address
									outer_loop_active = 1; //active the outer loop
								} //do nothing if the loop is already activated
								break;
							case 0x01: //inner loop header
								if(inner_loop_active == 0) {
									//Display "LooP" on the screen for 0.5 seconds
									spin(7); //wait 0.25 seconds 
									colon = 0; //colon/hyphen off
									azset(CONST_L,CONST_O,CONST_O,CONST_P,aTime);
									write_and_send(1);
									spin(15); //wait half a second

									inner_loop_register = M40+1; //number of iterations
									inner_loop_header = data_index; //record header address
									inner_loop_active = 1; //active the inner loop
								} //do nothing if the loop is already activated
								break;
						}
					
					} else { //IF1,IF0 are zero, so parse M4:M0 for meaning
						switch(M40) { 
							case 0: //Beginning of a routine
								colon = 0b00101000; //hyphen on
								//Display the routine number for 1 second then begin it
								//Find the routine number by matching data index to routine_addresses
								unsigned char i,r; //loop index and routine number
								r = 0; //start with routine #0 (nonvalid). should be 1..32 after search
								for(i=0;i<32;i++) {
									if(routine_addresses[i] == (data_index-1)) {
										r = i+1; //r is the routine number 1..32 to display
										break;
									}
								}							
								current_routine = r-1; //store for use in routine-selection instruction
								azset(CONST_OFF,CONST_R,r/10,r%10,aTime); //mins off, right 2 panels give routine number
								write_and_send(1);
								spin(31);
								break;
							case 1: //Display the clock in hours:minutes
								mode = PROGRAM_PAUSE; //flag that in paused mode
								colon = 0x82; //colon on
								while(mode == PROGRAM_PAUSE) {
									set_aTime_Hours_and_Minutes(); //get hrs:mins and write to aTime
									write_and_send(2);
									spin(63); //update only needs to happen every 2 seconds.
								}
								break;
							case 2: //Set clock (could be by user or by bluetooth cell phone's clock)
								//Display "Clock" as "CLoC" to indicate that user is setting the clock
								colon = 0; //colon off
								azset(CONST_C,CONST_L,CONST_O,CONST_C,aTime);
								write_and_send(1);
								spin(31);
								
								//Initialize the PCF chip to a default of 12:00 for setting time
								time_data_raw[1] = 0x12; //12 hrs
								time_data_raw[0] = 0x00; //00 mins
								PCF_register_setget(1,Hours,time_data_raw[1]);
								PCF_register_setget(1,Minutes,time_data_raw[0]);
								colon = 0x82; //colon on

								//Set the starting point for hours-selection
								set_aTime_Hours_and_Minutes(); //get hrs:mins and write to aTime
								hours_starting_point = 10*aTime[3]+aTime[2];
								minutes_starting_point = 10*aTime[1]+aTime[0];

								prompt_for_hours(hours_starting_point); //get user input for hours
								//Write the hours into the PCF8563 from final digit_selection
								digit_selection = ((digit_selection / 10) << 4) | (digit_selection % 10);
								PCF_register_setget(1, Hours, digit_selection);						
	
								prompt_for_minutes(minutes_starting_point); //get user input for mins
								//Write the minutes into the PCF8563 from final digit_selection
								digit_selection = ((digit_selection / 10) << 4) | (digit_selection % 10);
								PCF_register_setget(1, Minutes, digit_selection);
								break;
							case 3: //Prompt for routine number
								mode = ROUTINE_PROMPT; //flag that in routine-setting mode
								colon = 0b00101000; //hyphen on							

								char temp3; //nonvolatile variable used to read digit_selection
								digit_selection = current_routine+1; //# to display and cycle through (1..32)
								if(digit_selection > 32)
									digit_selection = 1; //loop around to routine #1
											
								while(mode == ROUTINE_PROMPT) {
									temp3 = digit_selection; //digit_selection modified over bluetooth
									//Display the user's current routine selection, blinking
									azset(CONST_OFF,11,temp3/10,temp3%10,aTime); //display OFF,"R",routine number
									write_and_send(1);
									spin(15); //wait half a second
									azset(CONST_OFF,CONST_R,CONST_OFF,CONST_OFF,aTime); //all panels off except routine "R"
									write_and_send(1);
									spin(7); //wait 0.25 seconds
								}
								break;
							case 4: //Schedule workout at time of day
								//Display "SCHL" as "5CHL" to indicate that user is setting the scheduled start time
								colon = 0; //colon off
								azset(5,CONST_C,CONST_H,CONST_L,aTime);
								write_and_send(1);
								spin(31);
								
								colon = 0x82; //colon on
								//Set the starting point for hours-selection
								set_aTime_Hours_and_Minutes(); //get hrs:mins and write to aTime
								hours_starting_point = 10*aTime[3]+aTime[2];
								minutes_starting_point = 0;
								
								//Use the top of the next hour as the initial setpoint
								if(hours_starting_point == 23)
									hours_starting_point = 0;
								else
									hours_starting_point++;

								prompt_for_hours(hours_starting_point); //get user input for hours
								hours_starting_point = digit_selection; //copy out user selection
	
								prompt_for_minutes(minutes_starting_point); //get user input for mins
								minutes_starting_point = digit_selection; //copy out user selection

								mode = SCHEDULED_START_MODE; //flag that in scheduled-start mode
								//Record the starting time in zTime[];
								set_azTime_Values(hours_starting_point,minutes_starting_point,zTime);
								set_aTime_Hours_and_Minutes(); //get hrs:mins and write to aTime

								//Until aTime reaches zTime, just display clock
								count_up = 1; //must be set to 1 for Tiems_Equal() to compare aTime[] to zTime[]
								while((mode == SCHEDULED_START_MODE) && (!(Times_Equal()))) { //display clock until reached scheduled time
									set_aTime_Hours_and_Minutes(); //get hrs:mins and write to aTime
									write_and_send(1);
									spin(63); //we don't need to update minutes every iteration
								}
								break;
							case 5: //Programmed Pause. Wait for user to advance.
								mode = PROGRAM_PAUSE; //flag that in paused mode
								time_data_raw[0] = PCF_register_setget(0, VL_seconds, 0); //get seconds
								time_data_raw[1] = PCF_register_setget(0, Minutes, 0); //get minutes
								blinks_done = 3;
								while(mode == PROGRAM_PAUSE) {
									//Keep clock chip from advancing by repeatedly writing same time:
									PCF_register_setget(1, VL_seconds, time_data_raw[0]);
									PCF_register_setget(1, Minutes, time_data_raw[1]);
						
									//Blink the colon to indicate the pause
									if(blinks_done == 0) { //toggle colon 8 times/sec (4Hz)
										colon ^= 0x82; //toggle the colon
										write_and_send(1);
										blinks_done = 3; //do not toggle again until 4 32Hz cycles later
									}
								}
								break;
							case 6: //Outer Conditional Loop-Back
								if(outer_loop_register > 0)
									outer_loop_register--; //decrement outer_loop register
								if(outer_loop_register == 0) //check if done iterating
									outer_loop_active = 0; //loop has completed->deactivate it
								else //more iterations exist, jump to loop header
									data_index = outer_loop_header;
								break;
							case 7: //Inner Conditional Loop-Back
								if(inner_loop_register > 0)
									inner_loop_register--; //decrement inner_loop register
								if(inner_loop_register == 0) //check if done iterating
									inner_loop_active = 0; //loop has completed->deactivate it
								else //more iterations exist, jump to loop header
									data_index = inner_loop_header;
								break;
							case 8: //Demo the screen
								colon = 0; //turn off colon/hyphen/decimals
								char j,k,m; //loop indeces
								for(j = 0; j < 2; j++) { //run through the demo 2 times
									for(k = 0; k < 11; k++) { //demo numbers 0..9 on the display, then all off
										azset(k,k,k,k,aTime); //write the digit to all panels
										write_and_send(0); //don't turn off upper panels for screen demo
										spin(7); //wait half a second
									}
									for(m = 0; m < 3; m++) {
										for(k = 1; k < 8; k+=2) {
											colon = (1<<k); //turn on colon/hyphen/period blocks 1 at a time
											write_and_send(0); //don't turn off upper panels for screen demo
											spin(3); //wait 0.125 seconds
										}
									}
									for(m = 0; m < 3; m++) {
										for(k = 0; k < 8; k+=2) {
											colon = (1<<k); //turn on colon/hyphen/period blocks 1 at a time
											write_and_send(0); //don't turn off upper panels for screen demo
											spin(3); //wait 0.125 seconds
										}
									}
								}
								break;
							case 9: //Display number of cycles remaining in innermost activated loop.
								if(inner_loop_active || outer_loop_active) { //display innermost active loop register
									spin(7); //wait 0.25 seconds to display last time before overwriting screen
									colon = 0b00101000; //hyphen on
									//Display (cycles remaining on innermost active loop), "C","R", 
									if(inner_loop_active) //display the inner loop register
										azset((inner_loop_register-1)/10,(inner_loop_register-1)%10,CONST_C,CONST_R,aTime);
									else //display the outer loop register
										azset((outer_loop_register-1)/10,(outer_loop_register-1)%10,CONST_C,CONST_R,aTime);
									write_and_send(2); //keep the ones-digit on, but tens can be turned off
									spin(15); //wait half a second
								}
								break;
							case 16: //Activate seconds: 10 milliseconds timer (must be timed by MCU, not PCF)
								mode = MILLISECOND_TIMER_MODE; //flag that in this mode

								//Zero out the time
								aTime[3] = 0; aTime[2] = 0; aTime[1] = 0; aTime[0] = 0;
								aTime_integer = 0;
								colon = 0x10; //turn on decimal place on panel 2

								GICR &= (~(1<<INT0)); //disable external interrupt INT0
								TIMSK = (1<<OCIE0); //Enable compare-match interupt on timer/counter0

								user_paused = 0;
								nextScreenReady = 0; //use to flag interaction with compare-match ISR
								while(mode == MILLISECOND_TIMER_MODE) { //count up in 10 milliseconds
									if(nextScreenReady == 0 ) { //if the next seconds.milliseconds display is not ready
										//Prepare the time
										//Write seconds and hundredths of seconds to aTime[] using aTime_integer
										set_azTime_Values((aTime_integer / 100) % 60,aTime_integer % 100,aTime);
										write_LED_buffer(aTime,colon,1);
										nextScreenReady = 1; //flag that the next thing to display is prepared
										
									}
								}
								
								TIMSK &= (~(1<<OCIE0)); //Disable compare-match interupt on timer/counter0

								//At this point, we have broken from millisecond timer and execution should continue
								nextScreenReady = 0; //must update screen for whatever follows this timing mode
								nextTimeReady = 0;
								GICR |= (1<<INT0); //enable external interrupt INT0
								break;
							default: //un parse-able. dump memory for analysis
								
								break;
						}
					}

				} else { //top bit is 0 -> timing format data
					count_up = 0x40 & byte0; //UD bit ==1->count up; ==0->count down
					colon = 0x82; //colon on, hyphen and decimal places off for timing intervals
					byte0 = 0x3F & byte0; //extract lower 6 bits for timing info (minutes)
					byte1 = 0x3F & get_data(data_index++); //read the second byte (seconds)

					//Parse the target time from the EEPROM 2-byte timing data
					zTime_integer = 60*((int)byte0) + byte1; //store time in seconds
					set_azTime_Values(byte0,byte1,zTime); //store target time
					//The target time stores the stored data whether counting up or down
					//For counting up, compare to target time. For counting down, compare to zero.					

					//Clear the PCF8563 minutes/seconds to begin it counting up from zero
					time_data_raw[1] = 0; time_data_raw[0] = 0;
					PCF_register_setget(1, VL_seconds, time_data_raw[0]);
					PCF_register_setget(1, Minutes, time_data_raw[1]);

					nextTimeReady = 1; //have parsed >=0 instructions and the next timing event
					//The next cycle through infinite loop will establish aTime. nextScreenReady is still 0
					do_not_update = 0; //timing data being executed: ISR can update the screen automatically
				} //end of if/else for determining instruction vs timing data
			} else { //only need to update the screen (no parsing required)
				set_aTime_Minutes_and_Seconds(); //set aTime with mins:secs data
				colon = 0x82; //turn on the colon
				//if counting up, no further calculation is necessary
				
				if(count_up == 0) { //counting down, so must calculate based on up-counting clock
					aTime_integer = (600 * ((int) aTime[3])) + (60 * ((int) aTime[2])) + (10 * ((int) aTime[1])) + ((int) aTime[0]);
					aTime_integer = zTime_integer-aTime_integer; //calculate remaining time in seconds
					//Get decimal encodings of mins,secs and write to aTime[]
					set_azTime_Values(aTime_integer / 60, aTime_integer % 60,aTime);
				}
				
				write_LED_buffer(aTime,colon,1); //write to the buffer with colon on
				nextScreenReady = 1; //we have updated the aTime for the next thing to display
			} //end of nextTimeReady==0 if/else statement
		
		} //end of nextScreenReady==0 if statement	
	} //end of the while(1) infinite loop
end //of main

//**********************************************************
/* Initialize all the peripherals and memory. */
void init()
begin
	//Configure the INT0 interrupt (connected to 32Hz PCF8563 output clock)
	MCUCR = ((1<<ISC01) | (1<<ISC00)); //issue interrupt on rising edge
	GICR = (1<<INT0); //enable INT0

	colon = 0x82; //colon on initially
	DDRD = 0; //inputs to avoid frying the bluetooth module

	//DO NOT USE THE SERIAL CONNECTION WHILE RN-42 IS CONNECTED!
	//Initialize the RS232 Interface
	uart_init();
  	stdout = stdin = stderr = &uart_str;
	fprintf(stdout,"Starting up...\n\r");

	//Initialize Bluetooth on USART
	//RN-42 module set to 9600baud, 8 bits, no parity, 1 stop bit, hardware flow control enabled
	//Parity can be turned on to even or odd through the RN-42 configurations

	UCSRB |= (1<<RXCIE); //enable the interrupt upon completed reception
/*
	//Turn on the receive interrupt enable, the receiver, the transmitter and 8bit frames
	UCSRB = ((1<<RXCIE) | (1<<RXEN) | (1<<TXEN));
	//Baud-rate high byte (value 0) is already set by default in the UCSRC address
	UBRRL = 64; //for 9600baud
	//8-bit Frame, No parity, 1 stop bit is already the default setting for UCSRC
*/

	i2c_init(); //initializes the TWI (I2C) interface
	init_PCF8563(); //initializes the clock chip
	LED_init(); //initializes the LED controllers and sends dot-correction data
	
	parse_routines(); //determine where the routine headers are
	ticks = 31; //initialize ticks to 1Hz cycle
	data_index = 0;
	user_paused = 0;

	//Some optimizations for lower power (ADC off by default).
	ACSR = (1<<ACD);  //turn off the analog comparator
	//Also disable the BODEN (brown-out detector) fuse

	//Idle sleep mode still allows the USART to wake the microcontroller,
	//which we do over Bluetooth.
	set_sleep_mode(SLEEP_MODE_IDLE); //configure the sleep mode for "Idle"
	sleep_enable(); //allow sleeping

	//Timer/Counter0 initialzation for millisecond timing functionality
	//Use CTC mode to clear timer upon compare match, and clk/1024 prescalar
	TCCR0 = ((1<<WGM01) | (1<<CS02) | (1<<CS00));
	OCR0 = 107; //108 ticks at clk/1024 gives 0.01 seconds for 11.0592MHz crystal

	mode = SLEEP_MODE; //turn on in sleep mode

	sei();	//start up the interrupts
end

//**********************************************************
/* Process User's "Prev" Button Press Based on Current State.
   This function passes on its argument k to parse_backwards. */
void Pressed_Prev_Button(char k)
begin
	switch(mode) {
		//Parse backwards for the following modes
		case SCHEDULED_START_MODE:
		case PROGRAM_PAUSE: //Parse backwards
		case NORMAL: //Parse backwards
		case USER_PAUSE: //unpause and Parse backwards
		case MILLISECOND_TIMER_MODE:
			parse_backwards(k);
			nextScreenReady = 0;
			nextTimeReady = 0;
			mode = NORMAL;
			break;
		case SET_CLOCK_HOURS: //decrement hours
			if(digit_selection == 0) //wrap around if at 00hrs
				digit_selection = 23;
			else
				digit_selection--;
			break;
		case SET_CLOCK_MINUTES: //decrement minutes
			if(digit_selection == 0) //wrap around if at 00mins
				digit_selection = 59;
			else
				digit_selection--;
			break;
		case ROUTINE_PROMPT://decrement routine number
			if(digit_selection == 1) //wrap around if at routine 1
				digit_selection = 32;
			else
				digit_selection--;
			break;
		case SLEEP_MODE: //this button does nothing in sleep mode
		default:
			break;
	} //end of switch statement
end

//**********************************************************
/* Process User's "Next" Button Press Based on Current State */
void Pressed_Next_Button()
begin
	switch(mode) {
		case SCHEDULED_START_MODE: //advance to next data
		case MILLISECOND_TIMER_MODE:
		case PROGRAM_PAUSE: //advance to next data
			mode = NORMAL; //fall through and parse next data
			break;
		case SET_CLOCK_HOURS: //increment hours
			if(digit_selection == 23) //wrap around if at 23hrs
				digit_selection = 0;
			else
				digit_selection++;
			break;
		case SET_CLOCK_MINUTES: //increment minutes
			if(digit_selection == 59) //wrap around if at 59mins
				digit_selection = 0;
			else
				digit_selection++;
			break;
		case ROUTINE_PROMPT://increment routine number
			if(digit_selection == 32) //wrap around if at routine 32
				digit_selection = 1;
			else
				digit_selection++;
			break;
		case NORMAL: //force it to parse the next instruction
			nextScreenReady = 0;
			nextTimeReady = 0;
			break;
		case USER_PAUSE: //unpause and force to parse the next instruction
			nextScreenReady = 0;
			nextTimeReady = 0;
			colon |= 0x82; //turn the colon back on if toggled odd # of times
			do_not_update = 0;
			mode = NORMAL;
			break;
		case SLEEP_MODE: //this button does nothing in sleep mode
		default:
			break;
	} //end of switch statement
end

//**********************************************************
/* Process User's "Enter" Button Press Based on Current State */
void Pressed_Enter_Button() 
begin
	switch(mode) {
		case SCHEDULED_START_MODE:  //enter does nothing in these modes
		case MILLISECOND_TIMER_MODE: 
		case NORMAL: 
		case PROGRAM_PAUSE: 
		case USER_PAUSE:
		case SLEEP_MODE:
			break;
		case SET_CLOCK_HOURS:
			mode = SET_CLOCK_MINUTES;
			break;
		case SET_CLOCK_MINUTES:
			mode = NORMAL;
			break;
		case ROUTINE_PROMPT:
			//Once user has selected the routine, go there:
			current_routine = digit_selection-1; //get the selected 0..31 routine index
			data_index = routine_addresses[current_routine]; //go to the address of that routine
			mode = NORMAL; //escape from loop upon hitting enter
			break;
		default:
			break;
	} //end of switch statement
end

//**********************************************************
/* Process User's "Play/Pause" Button Press Based on Current State */
void Pressed_Play_Pause_Button()
begin
	switch(mode) {
		case SCHEDULED_START_MODE: //play/pause does nothing in these modes
		case SET_CLOCK_HOURS:
		case SET_CLOCK_MINUTES:
		case SLEEP_MODE:
			break;
		case ROUTINE_PROMPT: //skip through routine prompt without jumping to another routine
			mode = NORMAL;
			break;
		case MILLISECOND_TIMER_MODE:
			if(user_paused)  //were already paused, so resume the ms-timer
				TIMSK |= (1<<OCIE0); //Enable compare-match interrupt on Timer/Counter0
			else //were not already paused, so pause the ms-timer
				TIMSK &= (~(1<<OCIE0)); //Disable compare-match interrupt on Timer/Counter0
			user_paused ^= 1; //toggle whether paused or not
			break;
		case PROGRAM_PAUSE: //hitting play/pause resumes
			mode = NORMAL;
			break;
		case NORMAL: //Pause the current timing event
			//Copy out the mins:secs time of the PCF
			time_data_raw[0] = PCF_register_setget(0, VL_seconds, 0); //get seconds
			time_data_raw[1] = PCF_register_setget(0, Minutes, 0); //get minutes
			mode = USER_PAUSE; //go into user-paused mode.
			do_not_update = 1; //the user-pause loop will update the screen.
			break;
		case USER_PAUSE: //unpause and resume normal operation
			colon |= 0x82; //turn the colon back on if toggled odd # of times
			do_not_update = 0;
			mode = NORMAL;
			break;
		default:
			break;
	} //end of switch statement
end

//**********************************************************
/* Process User's "Upload" Button Press Based on Current State */		
void Pressed_Upload_Button()
begin
	switch(mode) { //upload button does nothing in these modes
		case SLEEP_MODE:
		case UPLOAD_FAILED:
		case UPLOAD_SUCCESS:
		case UPLOAD:
			return;
			break;
		default:
			break;
	}

	//Go into upload mode by breaking any current loops and
	//wrapping around main infinite loop
	do_not_update = 1;
	upload_data_index = 0;
	mode = UPLOAD;
end

//**********************************************************
/* Process User's "Power" Button Press Based on Current State */	
void Pressed_Power_Button()
begin
	switch(mode) {
		case SLEEP_MODE:
			//Wake from sleep
			mode = NORMAL;
			reset_variables();
			break;
		case SCHEDULED_START_MODE:
		case SET_CLOCK_HOURS:
		case SET_CLOCK_MINUTES:
		case ROUTINE_PROMPT:
		case MILLISECOND_TIMER_MODE:
		case PROGRAM_PAUSE:
		case NORMAL:
		case USER_PAUSE:
			mode = SLEEP_MODE;
			do_not_update = 1; //the sleep-loop will handle the screen
		default:
			break;
	} //end of switch statement
end

//**********************************************************
/* Parses backwards through memory. Used to Process User's "Prev" Button Press.*/
void parse_backwards(char k)
begin
	char byte0, byte1; //used for reading the EEPROM memory (0,1 in memory index order)
	char j; //loop index that indicates how many entries to parse backwards

	//If it has been more than 1 second (32 ticks) since user pressed "prev"
	//then just restart the last executed instruction or interval
	
	for(j = 0; j < k; j++) { //loop to go back either one or two entries
		if(data_index == 0)
			return; //can't decrement below index of 0
		byte1 = get_data(--data_index); //get the last thing parsed
		if(0x80 & byte1) { //top bit is 1 -> instruction format data
			char IF10 = (0x03 & (byte1 >> 5)); //extract the IF1,IF0 bits
			unsigned char M40 = (0x1F & byte1);//extract the M4:M0 bits
			if(IF10) {
				switch(IF10) {
					case 0x03: //jump instruction
						parse_backwards(1); //skip back over the jump instruction
						break;
					case 0x02: //outer loop header
						outer_loop_active = 0; //restart loop with no iterations completed
						break;
					case 0x01: //inner loop header
						inner_loop_active = 0; //restart loop with no iterations completed
						break;
					default:
						break; //nothing else needs special attention
				}
			} else {
				if(M40 == 9)
					parse_backwards(1); //go back one more for ms-timer
			}
		} else { //last thing parsed was timing data, so go back one more byte
			byte0 = get_data(--data_index);
		} //end of instruction vs. timing data if/else statement
	} //end of for loop
	
	//Reset everything to parse the data from no initial conditions
	nextScreenReady = 0;
	nextTimeReady = 0;
	do_not_update = 0;
	mode = NORMAL;
end

//**********************************************************
/* Parse locations of the 32 routines. */
void parse_routines()
begin
	int i; //data_index variable
	unsigned char r;//routine index variable

	//clear all routine_addresses (there could be <32 routines)
	for(r = 0; r < 32; r++)
		routine_addresses[r] = 0;

	r = 0; //routines 1..32 store in r=0..31 of routine_addresses[r]
	for(i = 0; i < UPLOAD_SIZE; i++) {
		if (get_data(i) == 0x80) { //found a routine header
			routine_addresses[r] = i; //store routine header EEPROM index
			if(r < 32)
				r++; //avoid memory leaks by restricting r
		}
	}
end

//**********************************************************
/* Checks if the timing event is over by comparing aTime[] to zTime[] if counting up,
   or by comparing aTime[] to zero if counting down. Returns nonzero if interval is over. */
char Times_Equal()
begin
	if(count_up) //then compare aTime[] with the target time zTime[]
		return ((aTime[3] == zTime[3]) && (aTime[2] == zTime[2]) && (aTime[1] == zTime[1]) && (aTime[0] == zTime[0]));
	else
		return ((aTime[3] == 0) && (aTime[2] == 0) && (aTime[1] == 0) && (aTime[0] == 0));
end

//**********************************************************
/* Set the four values of array v to set_val3,set_val2,set_val1,set_val0.
   Functionally it is very simple, but saves 3 lines of code at each usage */
static inline void azset(char set_val3, char set_val2, char set_val1, char set_val0, unsigned char v[])
begin
	v[3] = set_val3;
	v[2] = set_val2;
	v[1] = set_val1;
	v[0] = set_val0;
end

//**********************************************************
/* Reads Hours and Minutes from the PCF8563, masks, and writes to aTime[] */
void set_aTime_Hours_and_Minutes()
begin
	time_data_raw[0] = PCF_register_setget(0, Minutes, 0); //get seconds
	time_data_raw[1] = PCF_register_setget(0, Hours, 0); //get minutes
	//Mask out the bits for ones/tens places according the PCF8563 Specification
	aTime[3] = (0x30 & time_data_raw[1])>>4; //hours tens-place
	aTime[2] = 0x0F & time_data_raw[1]; //hours ones-place
	aTime[1] = (0x70 & time_data_raw[0])>>4; //minutes tens-place
	aTime[0] = 0x0F & time_data_raw[0]; //minutes ones-place
end

//**********************************************************
/* Reads Hours and Minutes from the PCF8563, masks, and writes to aTime[] */
void set_aTime_Minutes_and_Seconds()
begin
	time_data_raw[0] = PCF_register_setget(0, VL_seconds, 0); //get seconds
	time_data_raw[1] = PCF_register_setget(0, Minutes, 0); //get minutes
	//Mask out the bits for ones/tens places according the PCF8563 Specification
	aTime[3] = (0x70 & time_data_raw[1])>>4; //minutes tens-place
	aTime[2] = 0x0F & time_data_raw[1]; //minutes ones-place
	aTime[1] = (0x70 & time_data_raw[0])>>4; //seconds tens-place
	aTime[0] = 0x0F & time_data_raw[0]; //seconds ones-place
end

//**********************************************************
/* Writes val1 (BCD) and val0 to array v (must have length 4).
   v[3] gets tens place of val1, v[1] gets tens place of val0.
   v[2] gets ones place of val1, v[0] gets ones place of val0. */
void set_azTime_Values(char val1, char val0, unsigned char v[])
begin
	v[3] = val1 / 10; //tens of larger time unit
	v[2] = val1 % 10; //ones of larger time unit
	v[1] = val0 / 10; //tens of smaller time unit
	v[0] = val0 % 10; //ones of smaller time unit
end

//**********************************************************
/* Performs the two common operations of writing to aTime then sending to LED drivers.
   If the null_high input ==1, then the upper two panels will be turned off as described
   in the write_LED_buffer function description, and if it ==2, just the upper one panel.*/
static inline void write_and_send(unsigned char null_high)
begin
	write_LED_buffer(aTime,colon,null_high); //write to the display buffer
	send_TLC5940(LED_buffer_long,96); //send to the LED drivers
end

//**********************************************************
/* Does Spin loop for "ticks_to_spin"+1 ticks at 32Hz rate (hardware counted).
   (31 ticks == 1 second)*/
static inline void spin(char ticks_to_spin)
begin
	blinks_done = ticks_to_spin; //# of 32Hz cycles until continuing (31 == 1 second)
	while(blinks_done != 0); //spin until display duration is over
end

//**********************************************************
/* Prompts user for hours input by flashing display and sitting
   in SET_CLOCK_HOURS mode. The resulting user input is stored
   in the global variable digit_selection. The argument is the
   starting point for digit_selection to display initially.*/
void prompt_for_hours(unsigned char starting_point_input)
begin
	mode = SET_CLOCK_HOURS; //flag that in clock-setting mode							
	digit_selection = starting_point_input;
	char temp2;			
	while(mode == SET_CLOCK_HOURS) {
		temp2 = digit_selection; //digit_selection modified over bluetooth
		//Display the user's current digit_selection, blinking
		azset(temp2/10,temp2%10,CONST_OFF,CONST_OFF,aTime); //hours displayed, mins off
		write_and_send(0);
		spin(15); //wait half a second
		azset(CONST_OFF,CONST_OFF,CONST_OFF,CONST_OFF,aTime); //all panels off
		write_and_send(0);
		spin(7); //wait 0.25 seconds
	}	
end

//**********************************************************
/* Prompts user for minutes input by flashing display and sitting
   in SET_CLOCK_MINUTES mode. Unlike prompt_for_hours, this
   function MUST be put into SET_CLOCK_MINUTES mode before executing
   for it to do anything. This is to prevent incorrect button-press
   response when leaving setting hours and advancing to setting
   minutes (clock-set and secheduled-start set modes). The resulting
   user input is stored in the global variable digit_selection. The
   argument is the starting point for digit_selection to display initially.*/
void prompt_for_minutes(unsigned char starting_point_input)
begin
	//mode = SET_CLOCK_MINUTES; MOVED TO THE ENTER_BUTTON_HANDLER
	digit_selection = starting_point_input;
	char temp2;
	while(mode == SET_CLOCK_MINUTES) {
		temp2 = digit_selection; //digit_selection modified over bluetooth
		//Display the user's current digit_selection, blinking
		azset(CONST_OFF,CONST_OFF,temp2/10,temp2%10,aTime); //hours off, right two panels=minutes
		write_and_send(0);
		spin(15); //wait half a second
		azset(CONST_OFF,CONST_OFF,CONST_OFF,CONST_OFF,aTime); //all panels off
		write_and_send(0);
		spin(7); //wait 0.25 seconds
	}
end

//**********************************************************
/* Resets variables for re-initialization after upload.*/
void reset_variables()
begin
	unsigned int i;
	for(i = 0; i < UPLOAD_SIZE; i++) 
		upload_buffer[i] = 0; //clear the upload buffer
	data_index = 0;
	inner_loop_active = 0;
	outer_loop_active = 0;
	current_routine = 0; 
	upload_data_index = 0; //start the beginning of the loop
	do_not_update = 1; //will be re-enabled when parsing timing interval
	nextScreenReady = 0;
	nextTimeReady = 0;
	GICR |= (1<<INT0); //enable external interrupt INT0
end

//**********************************************************
/* Hangs in user-pause mode until user escapes to another mode.
   Flashes the colon quickly (toggles 8 times/sec) to indicate
   that in a paused mode.*/
void handle_user_pause()
begin
	char first_entry;
	first_entry = 1; //used for more responsive pausing
	while(mode == USER_PAUSE) { //hang here until user resumes	
		do_not_update = 1;				
		//Keep clock chip from advancing by repeatedly writing same time:
		PCF_register_setget(1, VL_seconds, time_data_raw[0]);
		PCF_register_setget(1, Minutes, time_data_raw[1]);
		
		if(first_entry != 0) {//start flashing colon on first iteration
			first_entry = 0;
			blinks_done = 0;
		} 

		//Blink the colon to indicate the pause
		if(blinks_done == 0) { //toggle colon 8 times/sec (4Hz)
			colon ^= 0x82; //toggle the colon
			write_and_send(1);
			blinks_done = 3; //do not toggle again until 4 32Hz cycles later
		}
	} //do nothing if not in user-pause mode
	colon = 0x82; //turn colon back on if toggled odd # of times;
end


//**********************************************************
/* Handles settings for going to sleep, including disabling
   interrupts which we do not want waking the CPU and
   turning off the screen. Then, hangs in sleep mode until
   user escapes to another mode.*/
void handle_user_sleep()
begin
	if(mode == SLEEP_MODE) {
		do_not_update = 1;
		//Animate the screen turning off
		colon = 0x55;	azset(8,8,8,8,aTime);									write_and_send(0);	spin(15);
		colon = 0x15;	azset(CONST_OFF,8,8,8,aTime);							write_and_send(0);	spin(15);
		colon = 0x05;	azset(CONST_OFF,CONST_OFF,8,8,aTime); 					write_and_send(0);	spin(15);
		colon = 0x01;	azset(CONST_OFF,CONST_OFF,CONST_OFF,8,aTime);			write_and_send(0);	spin(15);
		colon = 0x00;	azset(CONST_OFF,CONST_OFF,CONST_OFF,CONST_OFF,aTime);	write_and_send(0);	spin(15);

		while(mode == SLEEP_MODE) {
			cli(); //disable interrupts
			//Leave only the USART interrupt on (out of the 3 ISRs used for this project)
			GICR &= (~(1<<INT0)); //Disable the INT0 interrupt from PCF8563 chip
			TIMSK &= (~(1<<OCIE0)); //Disable compare-match interupt on timer/counter0
	
			//Turn of the screen
			colon = 0; //colon off
			azset(CONST_OFF,CONST_OFF,CONST_OFF,CONST_OFF,aTime);
			write_and_send(0);

			sei();	//turn interrupts back on
			sleep_cpu(); //guaranteed to be executed b/c it follows sei();
			//Execution resumes here after handling the interrupt that wakes the CPU
		}

		//Animate the screen turning on
		colon = 0x00;	azset(CONST_OFF,CONST_OFF,CONST_OFF,CONST_OFF,aTime);	write_and_send(0);	spin(15);
		colon = 0x01;	azset(CONST_OFF,CONST_OFF,CONST_OFF,8,aTime);			write_and_send(0);	spin(15);
		colon = 0x05;	azset(CONST_OFF,CONST_OFF,8,8,aTime); 					write_and_send(0);	spin(15);
		colon = 0x15;	azset(CONST_OFF,8,8,8,aTime);							write_and_send(0);	spin(15);
		colon = 0x55;	azset(8,8,8,8,aTime);									write_and_send(0);	spin(15);

		reset_variables(); //upon waking up, read from start of data memory
		GICR |= (1<<INT0);//Re-enable the INT0 interrupt from PCF8563 chip
	} //do nothing if not in sleep mode
end

//**********************************************************
/* Prepares upload mode, then hangs in upload mode until user
   escapes to another mode or the upload timeout timer times out.
   Conveys success or failure of upload process on the screen.*/
void handle_user_upload()
begin
	//Handle all the uploading, parsing, and loading
	if(mode == UPLOAD) { //display this when entering upload mode
		do_not_update = 1;
		//Display "UPLo" to indicate in upload-mode
		colon = 0; //colon off
		azset(CONST_U,CONST_P,CONST_L,CONST_O,aTime);
		write_and_send(0);
		reset_variables();
		upload_timeout_timer = 0; //reset the timeout counter (counts up)

		while(mode == UPLOAD); 	//hang here until upload complete or timout occurs

		upload_data_index = 0; //reset index for writing to EEPROM
		//Ascertain success/failure of upload process, and respond based on it.
		colon = 0; //colon off
		if(mode == UPLOAD_SUCCESS) { //upload was a success -> write to the EEPROM
			//Overwrite the EEPROM
			for(upload_data_index = 0; upload_data_index < UPLOAD_SIZE; upload_data_index++) {
				//Uncomment to print the uploaded data
				//fprintf(stdout,"data_index[%d] will equal: %d\n\r",upload_data_index,upload_buffer[upload_data_index]);
				set_data(upload_buffer[upload_data_index], upload_data_index); //write uploaded bytes to EEPROM
			}
			azset(CONST_D,CONST_O,CONST_N,CONST_E,aTime); 	//Display "donE" for upload sucess
			parse_routines();
		} else { //upload was a failure
			azset(CONST_F,CONST_A,1,CONST_L,aTime); //Display "FA1L" for failed upload
		}
		//After (unsuccssfully or succesfully) uploading:
		write_and_send(0); //send the message to LED drivers
		reset_variables(); //start reading from beginning of data memory
		spin(63); //display success or failure message for 2 seconds
		mode = NORMAL;
	} //do nothing if not in upload mode
end
